"""
Provides a very thin wrapper to rethinkdb as well as common helper functions
"""
import rethinkdb
from .connection import connect
from .environment import check_log_env, check_prod_env, check_dev_env, log_env_gte, check_stage_env

r = rethinkdb #rethinkdb passthrough



"""
assortment of wrapped queries
"""
from ..brain_pb2 import Job
from ..checks import verify
from ..connection import rethinkdb as r
from .decorators import wrap_connection
from .decorators import wrap_rethink_generator_errors
from .decorators import wrap_rethink_errors
from . import RPX, RBT, RBJ, RBO, RPC, RPP


def _jobs_cursor(plugin_name):
    return RBJ.filter((r.row["JobTarget"]["PluginName"] == plugin_name) &
                      (r.row["Status"] == "Ready")).order_by('StartTime')


@wrap_rethink_generator_errors
@wrap_connection
def get_targets(conn=None):
    """
    get_brain_targets function from Brain.Targets table.

    :return: <generator> yields dict objects
    """
    targets = RBT
    results = targets.run(conn)
    for item in results:
        yield item


@wrap_rethink_generator_errors
@wrap_connection
def get_targets_by_plugin(plugin_name, conn=None):
    """
    get_targets_by_plugin function from Brain.Targets table

    :return: <generator> yields dict objects
    """
    targets = RBT
    results = targets.filter({"PluginName": plugin_name}).run(conn)
    for item in results:
        yield item


@wrap_rethink_generator_errors
@wrap_connection
def get_plugin_commands(plugin_name, conn=None):
    """
    get_specific_commands function queries Plugins.<PluginName> table
    the plugin name will be based off the user selection from the ui.
    It will return the query onto w2 as a dictionary nested in a list.

    :param plugin_name: <str> user's plugin selection
    :return: <generator> yields dictionaries
    """
    results = RPX.table(plugin_name).run(conn)
    for item in results:
        yield item


@wrap_rethink_errors
@wrap_connection
def get_plugin_command(plugin_name, command_name, conn=None):
    """
    get_specific_command function queries a specific CommandName

    :param plugin_name: <str> PluginName
    :param command_name: <str> CommandName
    :return: <dict>
    """
    commands = RPX.table(plugin_name).filter(
        {"CommandName": command_name}).run(conn)
    for command in commands:
        continue  # exhausting the cursor
    return command


@wrap_rethink_errors
@wrap_connection
def is_job_done(job_id, conn=None):
    """
    is_job_done function checks to if Brain.Jobs Status is 'Done'

    :param job_id: <str> id for the job
    :param conn: (optional)<connection> to run on
    :return: <dict> if job is done <false> if
    """
    result = False
    for item in RBJ.filter({'id': job_id,
                            'Status': "Done"}).run(conn):
        result = item
    return result


@wrap_rethink_errors
@wrap_connection
def get_output_content(job_id, max_size=1024, conn=None):
    """
    returns the content buffer for a job_id if that job output exists

    :param job_id: <str> id for the job
    :param max_size: <int> truncate after [max_size] bytes
    :param conn: (optional)<connection> to run on
    :return: <str> or <bytes>
    """
    content = None
    check_status = RBO.filter({"OutputJob": {'id': job_id}}).run(conn)
    for status_item in check_status:
        content = _truncate_output_content_if_required(status_item, max_size)
    return content


def _truncate_output_content_if_required(item, max_size):
    if max_size and "Content" in item and len(item['Content']) > max_size:
        content = "{}\n[truncated]".format(item['Content'][:max_size])
    elif "Content" in item:
        content = item['Content']
    else:
        content = ""
    return content


@wrap_rethink_errors
@wrap_connection
def plugin_exists(plugin_name, conn=None):
    """

    :param plugin_name:
    :param conn:
    :return: <bool> whether plugin exists
    """
    return plugin_name in RPX.table_list().run(conn)


@wrap_rethink_generator_errors
@wrap_connection
def get_jobs(plugin_name,
             verify_job=False, conn=None):
    """
    :param plugin_name: <str>
    :param verify_job: <bool>
    :param conn: <connection> or <NoneType>
    :return: <generator> yields <dict>
    """
    job_cur = _jobs_cursor(plugin_name).run(conn)
    for job in job_cur:
        if verify_job and not verify(job, Job()):
            continue #to the next job... warn?
        yield job


@wrap_rethink_errors
@wrap_connection
def get_next_job(plugin_name,
                 verify_job=False, conn=None):
    """

    :param plugin_name: <str>
    :param verify_job: <bool>
    :param conn: <connection> or <NoneType>
    :return: <dict> or <NoneType>
    """
    job_cur = _jobs_cursor(plugin_name).limit(1).run(conn)
    for job in job_cur:
        if verify_job and not verify(job, Job()):
            continue
        return job
    return None


@wrap_rethink_errors
@wrap_connection
def get_plugin_by_name_controller(plugin_name,
                                  conn=None):
    """

    :param plugin_name: <str> name of plugin
    :param conn: <rethinkdb.DefaultConnection>
    :return: <list> rethinkdb cursor
    """
    result = RPC.filter({
        "Name": plugin_name
    }).run(conn)
    return result


@wrap_rethink_errors
@wrap_connection
def get_ports_by_ip_controller(ip_address,
                               conn=None):
    """

    :param interface_name: <str> name of interface
    :param conn: <rethinkdb.DefaultConnection>
    :return: <list> rethinkdb cursor
    """
    if ip_address == "":
        result = RPP.filter({
            "Address": ip_address
        }).run(conn)
    else:
        result = RPP.filter(
            (r.row["Address"] == ip_address) |
            (r.row["Address"] == "")
        ).run(conn)
    return result



"""
requires fuse

this file is covered by test_binary_filesystem.py
but
it is spawned as a background thread,
so codeclimate doens't see the coverage

set it to nocover
"""

from sys import stderr
import stat
from multiprocessing import Lock
from time import time
from collections import defaultdict
from io import BytesIO

noop = lambda *args, **kwargs: None  # :pragma-nocover #PEP-559


try:  # :pragma-nocover
    from fuse import FUSE, FuseOSError, Operations, c_stat, ENOENT  # :pragma-nocover
    has_fuse = True  # :pragma-nocover
except ImportError as import_error:  # :pragma-nocover
    err_str = str(import_error)  # :pragma-nocover
    stderr.write("{1} - {0} requires fusepy\n".format(__name__, err_str))  # :pragma-nocover
    has_fuse = False  # :pragma-nocover
    FUSE = noop  # :pragma-nocover
    FuseOSError = None  # :pragma-nocover
    Operations = object  # :pragma-nocover
    c_stat = object  # :pragma-nocover

from .data import get, put, list_dir, delete
from .decorators import CONTENT_FIELD

VERBOSE = False
GET_DIR = [".", ".."]
ALLOW_LIST_DIR = True
ALLOW_REMOVE = False
READ_ONLY = False
MAX_CACHE_TIME = 600
OBJ_PERMISSION = 0o755


class NoStat(c_stat):  # pragma: no cover
    def __init__(self):  # pragma: no cover
        self.staged = False
        self.st_mode = 0
        self.st_ino = 0
        self.st_dev = 0
        self.st_nlink = 2
        self.st_uid = 0
        self.st_gid = 0
        self.st_size = 0
        self.st_atime = 0
        self.st_mtime = 0
        self.st_ctime = 0

    def as_dict(self):  # pragma: no cover
        return dict((key, getattr(self, key)) for key in ('st_atime',
                                                          'st_ctime',
                                                          'st_mtime',
                                                          'st_gid',
                                                          'st_mode',
                                                          'st_nlink',
                                                          'st_size',
                                                          'st_uid',
                                                          'st_ino'))


class BrainStore(Operations):
    """
    read only filesystem
    getattr should raise FuseOSError(ENOENT) when brain file doesn't exist
       and
    implement the create/write functions to be r/w
    """
    def __init__(self):
        self.cache = dict()
        self.attr = defaultdict(dict)
        self.attr_lock = Lock()

    def read(self, path, size, offset, fh):  # pragma: no cover
        # print("read {}".format(path))
        return self.cache[path][offset:offset+size]

    def readdir(self, path, fh):  # pragma: no cover
        # print("readdir {}".format(path))
        return GET_DIR + list_dir() if ALLOW_LIST_DIR else []

    def _getattr_root(self, base):  # pragma: no cover
        base.st_mode = int(stat.S_IFDIR | OBJ_PERMISSION)
        base.st_nlink = 2
        return base

    def _getattr_pull_file_to_cache(self, base, path):  # pragma: no cover
        filename = path.strip("/")
        brain_data = get(filename) or {}
        if not brain_data and not READ_ONLY:
            raise FuseOSError(ENOENT)
        buf = brain_data.get(CONTENT_FIELD, b"")
        base.st_mode = stat.S_IFREG | OBJ_PERMISSION
        base.st_nlink = 1
        base.st_size = len(buf)
        self.cache[path] = buf
        self.attr[path] = {"ts": time(), "base": base, "staged": None}
        return base

    def _getattr_file(self, base, path):  # pragma: no cover
        now_time = time()
        if now_time - self.attr[path].get("ts", 0) > MAX_CACHE_TIME:
            base = self._getattr_pull_file_to_cache(base, path)
        else:
            base = self.attr[path]['base']
        return base

    def getattr(self, path, fh=None):  # pragma: no cover
        # print("attr {}".format(path))
        base = NoStat()
        if path == "/":
            base = self._getattr_root(base)
        else:
            with self.attr_lock:
                base = self._getattr_file(base, path)
        return base.as_dict()

    def create(self, path, mode):  # pragma: no cover
        """
        This is currently a read-only filessytem.
        GetAttr will return a stat for everything
        if getattr raises FuseOSError(ENOENT)
        OS may call this function and the write function
        """
        # print("create {}".format(path))
        now_time = time()
        with self.attr_lock:
            base = NoStat()
            base.staged = True
            base.st_mode = stat.S_IFREG | OBJ_PERMISSION
            base.st_nlink = 1
            base.st_size = -1
            self.attr[path] = {"ts": now_time, "base": base, "staged": BytesIO()}
        return mode

    def write(self, path, data, offset, fh):  # pragma: no cover
        """
        This is a readonly filesystem right now
        """
        # print("write {}".format(path))
        with self.attr_lock:
            base = self.attr[path]['base']
            staged = self.attr[path]['staged']
            if not staged.closed:
                base.st_size += len(data)
                staged.write(data)
        return len(data)

    def unlink(self, path):  # pragma: no cover
        # print("unlink {}".format(path))
        with self.attr_lock:
            if path in self.attr:
                del self.cache[path]
                del self.attr[path]
                if ALLOW_REMOVE:
                    delete(path.strip("/"))

    def _release_upload_to_brain(self, path):  # pragma: no cover
        base = self.attr[path]['base']
        filename = path.strip("/")
        staged = self.attr[path]["staged"]
        if base.staged and base.st_size > 0 and not staged.closed:
            io_val = staged.getvalue()
            staged.close()
            try:
                put({"id": filename, "Name": filename, "Content": io_val})
            except ValueError as ValErr:
                stderr.write("{}\n".format(ValErr))
            del self.attr[path]

    def release(self, path, fh):  # pragma: no cover
        # print("release {}".format(path))
        with self.attr_lock:
            self._release_upload_to_brain(path)
        self._cleanup()
        return 0

    def _cleanup(self):  # pragma: no cover
        """
        cleans up data that's been in the cache for a while

        should be called from an async OS call like release? to not impact user
        :return:
        """
        need_to_delete = []  # can't delete from a dict while iterating
        with self.attr_lock:
            now_time = time()
            for path in self.cache:
                if now_time - self.attr[path]['ts'] >= MAX_CACHE_TIME:
                    need_to_delete.append(path)
            for path in need_to_delete:
                del self.attr[path]
                del self.cache[path]


def start_filesystem(mountpoint):  # pragma: no cover
    """
    prgramatically mount this filesystem to some mount point
    :param mountpoint:
    :return:
    """
    if has_fuse:
        FUSE(BrainStore(), mountpoint, foreground=True)
    else:
        raise ImportError(err_str)



"""
assortment of wrapped queries
"""
from ..brain_pb2 import Jobs, Target, Commands, Plugin, Port
from ..checks import verify
from ..connection import rethinkdb as r
from .decorators import wrap_connection
from .decorators import wrap_rethink_errors
from . import RPX, RBT, RBJ, RPC, RPP
from .reads import plugin_exists, get_plugin_by_name_controller, get_ports_by_ip_controller


def _check_port_conflict(port_data,
                         existing):
    for interface in existing:
        common_tcp = list(set(port_data["TCPPorts"]) &
                      set(interface["TCPPorts"]))
        if common_tcp != []:
            return {
                "errors": 1,
                "first_error": "TCP Port conflict(s): \
                {} in use on {}"
                .format(
                    common_tcp,
                    interface["Address"]
                )
            }
        common_udp = list(set(port_data["UDPPorts"]) &
                      set(interface["UDPPorts"]))
        if common_udp != []:
            return {
                "errors": 1,
                "first_error": "UDP Port conflict(s): \
                {} in use on {}"
                .format(
                    common_udp,
                    interface["Address"]
                )
            }
    return None


@wrap_rethink_errors
@wrap_connection
def insert_new_target(plugin_name, location_num,
                      port_num=0, optional="",
                      verify_target=False, conn=None):
    """
    insert_new_target function gets called when the user's input is validated
    and inserts a new target to Brain.Targets table.
    :param plugin_name: <str> user input plugin name
    :param location_num: <str> user input location number
    :param port_num: <str> user input port number
    :param optional: <str> user input optional
    :return: <dict> rethinkdb insert response value
    """
    target = {"PluginName": str(plugin_name),
              "Location": str(location_num),
              "Port": str(port_num),
              "Optional": {"init": str(optional)}}
    return insert_target(target, verify_target, conn)

@wrap_rethink_errors
@wrap_connection
def insert_target(target, verify_target=False,
                  conn=None):
    """

    :param target:  <dict> of Target() format
    :param verify_target: <bool>
    :param conn: <rethinkdb.DefaultConnection>
    :return: <dict> rethinkdb insert response value
    """
    if verify_target and not verify(target, Target()):
        raise ValueError("Invalid Target")
    return RBT.insert([target]).run(conn)

@wrap_rethink_errors
@wrap_connection
def insert_jobs(jobs, verify_jobs=True, conn=None):
    """
    insert_jobs function inserts data into Brain.Jobs table

    jobs must be in Job format

    :param jobs: <list> of Jobs
    :return: <dict> rethinkdb insert response value
    """
    assert isinstance(jobs, list)
    if verify_jobs and not verify({"Jobs": jobs}, Jobs()):
        raise ValueError("Invalid Jobs")
    inserted = RBJ.insert(jobs).run(conn)
    return inserted


@wrap_rethink_errors
@wrap_connection
def create_plugin(plugin_name, conn=None):
    """
    Creates a new plugin

    :param plugin_name: <str>
    :param conn:
    :return: <bool> successfully inserted
    """
    if not plugin_exists(plugin_name, conn=conn):
        RPX.table_create(plugin_name,
                         primary_key="CommandName"
                         ).run(conn)
    return True


@wrap_rethink_errors
@wrap_connection
def destroy_plugin(plugin_name, conn=None):
    """
    Creates a new plugin

    :param plugin_name: <str>
    :param conn:
    :return: <bool> successfully inserted
    """
    if plugin_exists(plugin_name, conn=conn):
        RPX.table_drop(plugin_name,
                       ).run(conn)
    return True


@wrap_rethink_errors
@wrap_connection
def advertise_plugin_commands(plugin_name, commands,
                              verify_commands=False,
                              conn=None):
    """

    :param plugin_name:
    :param commands:
    :param verify_commands:
    :param conn:
    :return:
    """
    assert isinstance(commands, list)
    if verify_commands and not verify({"Commands": commands},
                                      Commands()):
        raise ValueError("Invalid Commands")
    success = RPX.table(plugin_name).insert(commands,
                                            conflict="update"
                                            ).run(conn)
    return success


@wrap_rethink_errors
@wrap_connection
def create_plugin_controller(plugin_data,
                              verify_commands=False,
                              conn=None):
    """

    :param plugin_data: <dict> dict matching Plugin()
    :param verify_commands: <bool>
    :param conn: <rethinkdb.DefaultConnection>
    :return: <dict> rethinkdb insert response value
    """
    assert isinstance(plugin_data, dict)
    if verify_commands and not verify(plugin_data, Plugin()):
        raise ValueError("Invalid Plugin entry")
    current = get_plugin_by_name_controller(
        plugin_data["Name"],
        conn=conn
    )
    try:
        current.next()
        return {
            "errors": 1,
            "first_error": "".join([
                "Plugin ",
                plugin_data["Name"],
                " exists!"
            ])
        }
    except r.ReqlCursorEmpty:
        pass
    success = RPC.insert(plugin_data,
                         conflict="update"
                         ).run(conn)
    return success


@wrap_rethink_errors
@wrap_connection
def create_port_controller(port_data,
                           verify_port=False,
                           conn=None):
    """

    :param port_data: <dict> dict matching Port()
    :param verify_port: <bool>
    :param conn: <rethinkdb.DefaultConnection>
    :return: <dict> rethinkdb insert response value
    """
    assert isinstance(port_data, dict)
    if verify_port and not verify(port_data, Port()):
        raise ValueError("Invalid Port entry")
    existing = list(get_ports_by_ip_controller(
        port_data["Address"],
        conn=conn
    ))
    conflicts = _check_port_conflict(port_data, existing)
    if conflicts:
        return conflicts
    interface_existing = None
    for interface in existing:
        if interface["Address"] == port_data["Address"]:
            interface_existing = interface
    if not interface_existing:
        success = RPP.insert(
            port_data,
            conflict="update"
        ).run(conn)
    else:
        interface_existing["TCPPorts"] = list(set(port_data["TCPPorts"] + interface_existing["TCPPorts"]))
        interface_existing["UDPPorts"] = list(set(port_data["UDPPorts"] + interface_existing["UDPPorts"]))
        success = RPP.get(interface_existing["id"]).update({
            "TCPPorts": interface_existing["TCPPorts"],
            "UDPPorts": interface_existing["UDPPorts"]
        }).run(conn)
    return success


@wrap_rethink_errors
@wrap_connection
def update_plugin_controller(plugin_data,
                             verify_plugin=False,
                             conn=None):
    """

    :param plugin_data: <dict> dict matching Plugin()
    :param verify_commands: <bool>
    :param conn: <rethinkdb.DefaultConnection>
    :return: <dict> rethinkdb update response value
    """
    assert isinstance(plugin_data, dict)
    if verify_plugin and not verify(plugin_data, Plugin()):
        raise ValueError("Invalid Plugin entry")
    current = get_plugin_by_name_controller(
        plugin_data["Name"],
        conn=conn
    )
    update_id = None
    try:
        update_id = current.next()["id"]
    except r.ReqlCursorEmpty:
        return {
            "errors": 1,
            "first_error": "Cannot update non-existent plugin!"
        }
    success = RPC.get(update_id).update(plugin_data).run(conn)
    return success



"""
decorators fro the binary module
"""
from decorator import decorator
from .. import r
# import magic at bottom of file

BINARY = r.binary
PRIMARY_FIELD = "Name"
PRIMARY_KEY = "id"
CONTENT_FIELD = "Content"
CONTENTTYPE_FIELD = "ContentType"


@decorator
def wrap_name_to_id(func_, *args, **kwargs):
    """
    destination (rethinkdb) needs the id field as primary key
    put the Name field into the ID field
    :param func_:
    :param args:
    :param kwargs:
    :return:
    """
    assert isinstance(args[0], dict)
    args[0][PRIMARY_KEY] = args[0].get(PRIMARY_FIELD, "")
    return func_(*args, **kwargs)


@decorator
def wrap_guess_content_type(func_, *args, **kwargs):
    """
    guesses the content type with libmagic if available
    :param func_:
    :param args:
    :param kwargs:
    :return:
    """
    assert isinstance(args[0], dict)
    if not args[0].get(CONTENTTYPE_FIELD, None):
        content = args[0].get(CONTENT_FIELD, b"")
        try:
            args[0][CONTENTTYPE_FIELD] = magic.from_buffer(content)
        except magic.MagicException:  # pragma: no cover
            args[0][CONTENTTYPE_FIELD] = MockMagic.DEFAULT_MAGIC
    return func_(*args, **kwargs)


@decorator
def wrap_content_as_binary_if_needed(func_, *args, **kwargs):
    """
    destination (rethinkdb) needs the id field as primary key
    put the Name field into the ID field
    :param func_:
    :param args:
    :param kwargs:
    :return:
    """
    assert isinstance(args[0], dict)
    try:
        args[0][CONTENT_FIELD] = BINARY(args[0].get(CONTENT_FIELD, b""))
    except AttributeError:  # pragma: no cover
        pass  # toss in the object as string
    return func_(*args, **kwargs)


class MockMagic(object):  # pragma: no cover
    DEFAULT_MAGIC = "data"

    class MagicException(Exception):  # pragma: no cover
        pass

    @staticmethod
    def from_buffer(content=None):  # pragma: no cover
        return MockMagic.DEFAULT_MAGIC


try:
    import magic
except ImportError:  # pragma: no cover
    magic = MockMagic



"""
controls Job related changes
"""
from decorator import decorator
from .brain_pb2 import Job, Jobs
from .queries import insert_jobs, is_job_done, get_next_job
from .checks import verify

BEGIN = ""
INVALID = "Invalid"
VALID = "Valid"
READY = "Ready"
STOP = "Stopped"
PENDING = "Pending"
DONE = "Done"
ERROR = "Error"
WAITING = "Waiting"
ACTIVE = "Active"

SUCCESS = "success"
FAILURE = "failure"
TRANSITION = "transition"

class JobsError(Exception):
    """
    Base Jobs exception
    """


class InvalidStateTransition(JobsError):
    """
    Simple exception to identify invalid state transition

    State Transition governed by The Brain Documentation
    """
    pass


class InvalidState(JobsError):
    """
    Simple exception to identify invalid state transition

    State Transition governed by The Brain Documentation
    """
    pass


STATES = {BEGIN: {SUCCESS: READY,
                  FAILURE: ERROR,
                  TRANSITION: frozenset([VALID,
                                         READY,
                                         INVALID,
                                         ERROR,
                                         STOP])},
          INVALID: {SUCCESS: VALID,
                    FAILURE: INVALID,
                    TRANSITION: frozenset([VALID,
                                           INVALID])},
          VALID: {SUCCESS: READY,
                  FAILURE: INVALID,
                  TRANSITION: frozenset([READY,
                                         WAITING,
                                         INVALID])},
          READY: {SUCCESS: PENDING,
                  FAILURE: ERROR,
                  TRANSITION: frozenset([STOP,
                                         PENDING,
                                         ERROR])},
          PENDING: {SUCCESS: DONE,
                    FAILURE: ERROR,
                    TRANSITION: frozenset([STOP,
                                           ERROR,
                                           ACTIVE,
                                           DONE])},
          DONE: {SUCCESS: DONE,
                 FAILURE: ERROR,
                 TRANSITION: frozenset([DONE,
                                        ERROR])},
          WAITING: {SUCCESS: READY,
                    FAILURE: ERROR,
                    TRANSITION: frozenset([READY,
                                           ERROR,
                                           STOP])},
          ACTIVE: {SUCCESS: DONE,
                   FAILURE: ERROR,
                   TRANSITION: frozenset([ERROR,
                                          DONE])},
          STOP: {SUCCESS: STOP,
                 FAILURE: ERROR,
                 TRANSITION: frozenset([STOP,
                                        ERROR,
                                        WAITING,
                                        READY])},
          ERROR: {SUCCESS: DONE,
                  FAILURE: ERROR,
                  TRANSITION: frozenset([DONE,
                                         ERROR,
                                         STOP])}
          }


@decorator
def wrap_good_state(f, *args, **kwargs):
    """
    Decorator/Wrapper to verify the input is an acceptable state
    prior to calling a function on it

    :param f: <function>  to call
    :param args: <tuple> positional args
    :param kwargs: <dict> keyword args
    :return: wrapped function return
    """
    if not verify_state(args[0]):
        raise InvalidState("{} is not a valid state".format(args[0]))
    return f(*args, **kwargs)


def verify_state(state):
    """
    Verifies a state is acceptable

    :param state: <str> state name
    :return: <bool>
    """
    return state in STATES


@wrap_good_state
def transition_success(state):
    """
    transition to the on-success state

    :param state: <str> state name
    :return: <str> or None
    """
    return STATES[state][SUCCESS]


@wrap_good_state
def transition_fail(state):
    """
    transition to the on-failure state

    :param state: <str> state name
    :return: <str> or None
    """
    return STATES[state][FAILURE]


@wrap_good_state
def transition(prior_state, next_state):
    """
    Transitions to a non-standard state

    Raises InvalidStateTransition if next_state is not allowed.

    :param prior_state: <str>
    :param next_state: <str>
    :return: <str>
    """
    if next_state not in STATES[prior_state][TRANSITION]:
        acceptable = STATES[prior_state][TRANSITION]
        err = "cannot {}->{} may only {}->{}".format(prior_state,
                                                     next_state,
                                                     prior_state,
                                                     acceptable)
        raise InvalidStateTransition(err)
    return next_state


def verify_job(job):
    """
    verifies a job

    :param job: <dict>
    :return: <bool>
    """
    return verify(job, Job())



#!/usr/bin/python
from setuptools import setup, find_packages

VERSION = "0.1.36"

setup(
    name="ramrodbrain",
    version=VERSION,
    packages=find_packages(),
    description='RethinkDB Wrapper functions',
    author='Dan Bauman',
    maintainer='Dan Bauman',
    maintainer_email='dan@bauman.space',
    license='MIT',
    url='https://github.com/ramrod-project/database-brain',
    download_url='https://github.com/ramrod-project/python-brain/archive/ramrodbrain-{}.tar.gz'.format(VERSION),
    install_requires=[
        'rethinkdb',
        "protobuf",
        "dict-to-protobuf",
        "decorator",
        "python-magic"
    ],
    classifiers=[
                       'License :: OSI Approved :: MIT License',
                       'Operating System :: OS Independent',
                       'Programming Language :: Python',
                       'Programming Language :: Python :: 3',
                   ],
)


"""
functions to perform checks used by other areas of the brain

functions should all return a boolean if possible.
"""

from google.protobuf.message import EncodeError
import dict_to_protobuf #this lib allows extra keys
dict_to_protobuf.l.setLevel("ERROR")


def verify(value, msg):
    """
    C-style validator

    Keyword arguments:
    value -- dictionary to validate (required)
    msg -- the protobuf schema to validate against (required)

    Returns:
        True: If valid input
        False: If invalid input
    """
    return converts_to_proto(value, msg) and \
           successfuly_encodes(msg) and \
           special_typechecking(value, msg)


def special_typechecking(value, msg):
    """
    Special Typechecking not available via protocol itself
    :param value: <dict>
    :param msg: <proto object>
    :return: <bool>
    """
    result = True
    if msg.DESCRIPTOR.name == "Target":
        result &= special_target_typecheck(value)
    return result


def special_target_typecheck(value):
    """
    Special type checking for the target object
    :param value: <dict>
    :return: <bool>
    """
    result = True
    # if key:Optional exists, it must be a dict object
    result &= isinstance(value.get("Optional", dict()), dict)
    return result


def converts_to_proto(value, msg, raise_err=False):
    """
    Boolean response if a dictionary can convert into the proto's schema

    :param value: <dict>
    :param msg: <proto object>
    :param raise_err: <bool> (default false) raise for troubleshooting
    :return: <bool> whether the dict can covert
    """
    result = True
    try:
        dict_to_protobuf.dict_to_protobuf(value, msg)
    except TypeError as type_error:
        if raise_err:
            raise type_error
        result = False
    return result


def successfuly_encodes(msg, raise_err=False):
    """
    boolean response if a message contains correct information to serialize

    :param msg: <proto object>
    :param raise_err: <bool>
    :return: <bool>
    """
    result = True
    try:
        msg.SerializeToString()
    except EncodeError as encode_error:
        if raise_err:
            raise encode_error
        result = False
    return result


def strip(value, msg):
    """
    Strips all non-essential keys from the value dictionary
    given the message format protobuf

    raises ValueError exception if value does not have all required keys

    :param value: <dict> with arbitrary keys
    :param msg: <protobuf> with a defined schema
    :return: NEW <dict> with keys defined by msg, omits any other key
    """
    dict_to_protobuf.dict_to_protobuf(value, msg)
    try:
        msg.SerializeToString()  #raise error for insufficient input
    except EncodeError as encode_error:
        raise ValueError(str(encode_error))
    output = dict_to_protobuf.protobuf_to_dict(msg)
    return output
